module Giraffe.Htmx

open Microsoft.AspNetCore.Http
open Microsoft.Extensions.Primitives
open System

/// Determine if the given header is present
let private hdr (headers : IHeaderDictionary) hdr =
    match headers[hdr] with it when it = StringValues.Empty -> None | it -> Some it[0]

/// Extensions to the header dictionary
type IHeaderDictionary with
  
    /// <summary>Indicates that the request is via an element using <tt>hx-boost</tt></summary>
    member this.HxBoosted
        with get () = hdr this "HX-Boosted" |> Option.map bool.Parse
    
    /// <summary>The current URL of the browser <em>(note that this does not update until after settle)</em></summary>
    member this.HxCurrentUrl
        with get () = hdr this "HX-Current-URL" |> Option.map Uri
    
    /// <summary>
    /// <tt>true</tt> if the request is for history restoration after a miss in the local history cache
    /// </summary>
    member this.HxHistoryRestoreRequest
        with get () = hdr this "HX-History-Restore-Request" |> Option.map bool.Parse
    
    /// <summary>The user response to an <tt>hx-prompt</tt></summary>
    member this.HxPrompt
        with get () = hdr this "HX-Prompt"

    /// <summary><tt>true</tt> if the request came from htmx</summary>
    member this.HxRequest
        with get () = hdr this "HX-Request" |> Option.map bool.Parse

    /// <summary>The <tt>id</tt> attribute of the target element if it exists</summary>
    member this.HxTarget
        with get () = hdr this "HX-Target"

    /// <summary>The <tt>id</tt> attribute of the triggered element if it exists</summary>
    member this.HxTrigger
        with get () = hdr this "HX-Trigger"

    /// <summary>The <tt>name</tt> attribute of the triggered element if it exists</summary>
    member this.HxTriggerName
        with get () = hdr this "HX-Trigger-Name"


/// Extensions for the request object
type HttpRequest with

    /// <summary>Whether this request was initiated from htmx</summary>
    member this.IsHtmx
        with get () = this.Headers.HxRequest |> Option.defaultValue false

    /// <summary>Whether this request is an htmx history-miss refresh request</summary>
    member this.IsHtmxRefresh
        with get () = this.IsHtmx && (this.Headers.HxHistoryRestoreRequest |> Option.defaultValue false)


/// <summary>HTTP handlers for setting output headers</summary>
[<AutoOpen>]
module Handlers =

    open Giraffe.Htmx.Common
    
    /// <summary>Pushes a new url into the history stack</summary>
    /// <param name="url">The URL to be pushed</param>
    /// <returns>An HTTP handler with the <tt>HX-Push-Url</tt> header set</returns>
    let withHxPushUrl (url: string) : HttpHandler =
        setHttpHeader "HX-Push-Url" url

    /// <summary>Explicitly do not push a new URL into the history stack</summary>
    /// <returns>An HTTP handler with the <tt>HX-Push-Url</tt> header set to <tt>false</tt></returns>
    let withHxNoPushUrl : HttpHandler =
        toLowerBool false |> withHxPushUrl
      
    /// Pushes a new url into the history stack
    [<Obsolete "Use withHxPushUrl; HX-Push was replaced by HX-Push-Url in v1.8.0">]
    let withHxPush = withHxPushUrl
    
    /// Explicitly do not push a new URL into the history stack
    [<Obsolete "Use withHxNoPushUrl; HX-Push was replaced by HX-Push-Url in v1.8.0">]
    let withHxNoPush = withHxNoPushUrl
      
    /// <summary>Can be used to do a client-side redirect to a new location</summary>
    /// <param name="url">The URL to which the client should be redirected</param>
    /// <returns>An HTTP handler with the <tt>HX-Redirect</tt> header set</returns>
    let withHxRedirect (url: string) : HttpHandler =
        setHttpHeader "HX-Redirect" url

    /// <summary>If set to <tt>true</tt> the client side will do a full refresh of the page</summary>
    /// <param name="shouldRefresh">Whether the client should refresh their page</param>
    /// <returns>An HTTP handler with the <tt>HX-Refresh</tt> header set</returns>
    let withHxRefresh shouldRefresh : HttpHandler =
        (toLowerBool >> setHttpHeader "HX-Refresh") shouldRefresh

    /// <summary>Replaces the current URL in the history stack</summary>
    /// <param name="url">The URL to place in the history stack in place of the current one</param>
    /// <returns>An HTTP handler with the <tt>HX-Replace-URL</tt> header set</returns>
    let withHxReplaceUrl url : HttpHandler =
        setHttpHeader "HX-Replace-Url" url

    /// <summary>Explicitly do not replace the current URL in the history stack</summary>
    /// <returns>An HTTP handler with the <tt>HX-Replace-URL</tt> header set to <tt>false</tt></returns>
    let withHxNoReplaceUrl : HttpHandler =
        toLowerBool false |> withHxReplaceUrl
    
    /// <summary>Override which portion of the response will be swapped into the target document</summary>
    /// <param name="target">The selector for the new response target</param>
    /// <returns>An HTTP handler with the <tt>HX-Reselect</tt> header set</returns>
    let withHxReselect (target: string) : HttpHandler =
        setHttpHeader "HX-Reselect" target
    
    /// <summary>Override the <tt>hx-swap</tt> attribute from the initiating element</summary>
    /// <param name="swap">The swap value to override</param>
    /// <returns>An HTTP handler with the <tt>HX-Reswap</tt> header set</returns>
    /// <remarks>Use <see cref="T:Giraffe.Htmx.Common.HxSwap">HxSwap</see> constants for best results</remarks>
    let withHxReswap (swap: string) : HttpHandler =
        setHttpHeader "HX-Reswap" swap

    /// <summary>Allows you to override the <tt>hx-target</tt> attribute</summary>
    /// <param name="target">The new target for the response</param>
    /// <returns>An HTTP handler with the <tt>HX-Retarget</tt> header set</returns>
    let withHxRetarget (target: string) : HttpHandler =
        setHttpHeader "HX-Retarget" target

    /// <summary>Allows you to trigger a single client side event</summary>
    /// <param name="evt">The call to the event that should be triggered</param>
    /// <returns>An HTTP handler with the <tt>HX-Trigger</tt> header set</returns>
    let withHxTrigger (evt: string) : HttpHandler =
        setHttpHeader "HX-Trigger" evt

    /// <summary>Allows you to trigger multiple client side events</summary>
    /// <param name="evts">The calls to events that should be triggered</param>
    /// <returns>An HTTP handler with the <tt>HX-Trigger</tt> header set for all given events</returns>
    let withHxTriggerMany evts : HttpHandler =
        toJson evts |> setHttpHeader "HX-Trigger"

    /// <summary>Allows you to trigger a single client side event after changes have settled</summary>
    /// <param name="evt">The call to the event that should be triggered</param>
    /// <returns>An HTTP handler with the <tt>HX-Trigger-After-Settle</tt> header set</returns>
    let withHxTriggerAfterSettle (evt: string) : HttpHandler =
        setHttpHeader "HX-Trigger-After-Settle" evt

    /// <summary>Allows you to trigger multiple client side events after changes have settled</summary>
    /// <param name="evts">The calls to events that should be triggered</param>
    /// <returns>An HTTP handler with the <tt>HX-Trigger-After-Settle</tt> header set for all given events</returns>
    let withHxTriggerManyAfterSettle evts : HttpHandler =
        toJson evts |> setHttpHeader "HX-Trigger-After-Settle"

    /// <summary>Allows you to trigger a single client side event after DOM swapping occurs</summary>
    /// <param name="evt">The call to the event that should be triggered</param>
    /// <returns>An HTTP handler with the <tt>HX-Trigger-After-Swap</tt> header set</returns>
    let withHxTriggerAfterSwap (evt: string) : HttpHandler =
        setHttpHeader "HX-Trigger-After-Swap" evt

    /// <summary>Allows you to trigger multiple client side events after DOM swapping occurs</summary>
    /// <param name="evts">The calls to events that should be triggered</param>
    /// <returns>An HTTP handler with the <tt>HX-Trigger-After-Swap</tt> header set for all given events</returns>
    let withHxTriggerManyAfterSwap evts : HttpHandler =
        toJson evts |> setHttpHeader "HX-Trigger-After-Swap"
