/// <summary>Types and functions supporting htmx attributes in Giraffe View Engine</summary>
module Giraffe.ViewEngine.Htmx

/// <summary>Valid values for the <tt>hx-encoding</tt> attribute</summary>
[<RequireQualifiedAccess>]
module HxEncoding =
    
    /// <summary>A standard HTTP form</summary>
    let Form = "application/x-www-form-urlencoded"
    
    /// <summary>A multipart form (used for file uploads)</summary>
    let MultipartForm = "multipart/form-data"


/// <summary>The events recognized by htmx</summary>
[<Struct>]
type HxEvent =
    
    /// <summary>Send this event to an element to abort a request</summary>
    | Abort
    
    /// <summary>Triggered after an AJAX request has completed processing a successful response</summary>
    | AfterOnLoad
    
    /// <summary>Triggered after htmx has initialized a node</summary>
    | AfterProcessNode
    
    /// <summary>Triggered after an AJAX request has completed</summary>
    | AfterRequest
    
    /// <summary>Triggered after the DOM has settled</summary>
    | AfterSettle
    
    /// <summary>Triggered after new content has been swapped in</summary>
    | AfterSwap
    
    /// <summary>Triggered before htmx disables an element or removes it from the DOM</summary>
    | BeforeCleanupElement
    
    /// <summary>Triggered before any response processing occurs</summary>
    | BeforeOnLoad
    
    /// <summary>Triggered before htmx initializes a node</summary>
    | BeforeProcessNode
    
    /// <summary>Triggered before an AJAX request is made</summary>
    | BeforeRequest
    
    /// <summary>Triggered before a swap is done, allows you to configure the swap</summary>
    | BeforeSwap
    
    /// <summary>Triggered just before an ajax request is sent</summary>
    | BeforeSend
    
    /// <summary>Triggered before the request, allows you to customize parameters, headers</summary>
    | ConfigRequest
    
    /// <summary>
    /// Triggered after a trigger occurs on an element, allows you to cancel (or delay) issuing the AJAX request
    /// </summary>
    | Confirm
    
    /// <summary>Triggered on an error during cache writing</summary>
    | HistoryCacheError
    
    /// <summary>Triggered on a cache miss in the history subsystem</summary>
    | HistoryCacheMiss
    
    /// <summary>Triggered on a unsuccessful remote retrieval</summary>
    | HistoryCacheMissError
    
    /// <summary>Triggered on a successful remote retrieval</summary>
    | HistoryCacheMissLoad
    
    /// <summary>Triggered when htmx handles a history restoration action</summary>
    | HistoryRestore
    
    /// <summary>Triggered before content is saved to the history cache</summary>
    | BeforeHistorySave
    
    /// <summary>Triggered when new content is added to the DOM</summary>
    | Load
    
    /// <summary>
    /// Triggered when an element refers to a SSE event in its trigger, but no parent SSE source has been defined
    /// </summary>
    | NoSseSourceError
    
    /// <summary>Triggered when an exception occurs during the onLoad handling in htmx</summary>
    | OnLoadError
    
    /// <summary>Triggered after an out of band element as been swapped in</summary>
    | OobAfterSwap
    
    /// <summary>Triggered before an out of band element swap is done, allows you to configure the swap</summary>
    | OobBeforeSwap
    
    /// <summary>Triggered when an out of band element does not have a matching ID in the current DOM</summary>
    | OobErrorNoTarget
    
    /// <summary>Triggered after a prompt is shown</summary>
    | Prompt
    
    /// <summary>Triggered after an url is pushed into history</summary>
    | PushedIntoHistory
    
    /// <summary>Triggered when an HTTP response error (non-200 or 300 response code) occurs</summary>
    | ResponseError
    
    /// <summary>Triggered when a network error prevents an HTTP request from happening</summary>
    | SendError
    
    /// <summary>Triggered when an error occurs with a SSE source</summary>
    | SseError
    
    /// <summary>Triggered when a SSE source is opened</summary>
    | SseOpen
    
    /// <summary>Triggered when an error occurs during the swap phase</summary>
    | SwapError
    
    /// <summary>Triggered when an invalid target is specified</summary>
    | TargetError
    
    /// <summary>Triggered when a request timeout occurs</summary>
    | Timeout
    
    /// <summary>Triggered before an element is validated</summary>
    | ValidationValidate
    
    /// <summary>Triggered when an element fails validation</summary>
    | ValidationFailed
    
    /// <summary>Triggered when a request is halted due to validation errors</summary>
    | ValidationHalted
    
    /// <summary>Triggered when an ajax request aborts</summary>
    | XhrAbort
    
    /// <summary>Triggered when an ajax request ends</summary>
    | XhrLoadEnd
    
    /// <summary>Triggered when an ajax request starts</summary>
    | XhrLoadStart
    
    /// <summary>Triggered periodically during an ajax request that supports progress events</summary>
    | XhrProgress
    
    /// The htmx event name (fst) and kebab-case name (snd, for use with hx-on)
    static member private Values = Map [
        Abort, ("abort", "abort")
        AfterOnLoad, ("afterOnLoad", "after-on-load")
        AfterProcessNode, ("afterProcessNode", "after-process-node")
        AfterRequest, ("afterRequest", "after-request")
        AfterSettle, ("afterSettle", "after-settle")
        AfterSwap, ("afterSwap", "after-swap")
        BeforeCleanupElement, ("beforeCleanupElement", "before-cleanup-element")
        BeforeOnLoad, ("beforeOnLoad", "before-on-load")
        BeforeProcessNode, ("beforeProcessNode", "before-process-node")
        BeforeRequest, ("beforeRequest", "before-request")
        BeforeSwap, ("beforeSwap", "before-swap")
        BeforeSend, ("beforeSend", "before-send")
        ConfigRequest, ("configRequest", "config-request")
        Confirm, ("confirm", "confirm")
        HistoryCacheError, ("historyCacheError", "history-cache-error")
        HistoryCacheMiss, ("historyCacheMiss", "history-cache-miss")
        HistoryCacheMissError, ("historyCacheMissError", "history-cache-miss-error")
        HistoryCacheMissLoad, ("historyCacheMissLoad", "history-cache-miss-load")
        HistoryRestore, ("historyRestore", "history-restore")
        BeforeHistorySave, ("beforeHistorySave", "before-history-save")
        Load, ("load", "load")
        NoSseSourceError, ("noSSESourceError", "no-sse-source-error")
        OnLoadError, ("onLoadError", "on-load-error")
        OobAfterSwap, ("oobAfterSwap", "oob-after-swap")
        OobBeforeSwap, ("oobBeforeSwap", "oob-before-swap")
        OobErrorNoTarget, ("oobErrorNoTarget", "oob-error-no-target")
        Prompt, ("prompt", "prompt")
        PushedIntoHistory, ("pushedIntoHistory", "pushed-into-history")
        ResponseError, ("responseError", "response-error")
        SendError, ("sendError", "send-error")
        SseError, ("sseError", "sse-error")
        SseOpen, ("sseOpen", "sse-open")
        SwapError, ("swapError", "swap-error")
        TargetError, ("targetError", "target-error")
        Timeout, ("timeout", "timeout")
        ValidationValidate, ("validation:validate", "validation:validate")
        ValidationFailed, ("validation:failed", "validation:failed")
        ValidationHalted, ("validation:halted", "validation:halted")
        XhrAbort, ("xhr:abort", "xhr:abort")
        XhrLoadEnd, ("xhr:loadend", "xhr:loadend")
        XhrLoadStart, ("xhr:loadstart", "xhr:loadstart")
        XhrProgress, ("xhr:progress", "xhr:progress")
    ]
    
    /// <summary>The htmx event name</summary> 
    override this.ToString() = fst HxEvent.Values[this]
    
    /// <summary>The <tt>hx-on</tt> variant of the htmx event name</summary>
    member this.ToHxOnString() = snd HxEvent.Values[this]


/// <summary>Helper to create the <tt>hx-headers</tt> attribute</summary>
[<RequireQualifiedAccess>]
module HxHeaders =
    
    /// <summary>Create headers from a list of key/value pairs</summary>
    let From = Giraffe.Htmx.Common.toJson


/// <summary>Values / helpers for the <tt>hx-params</tt> attribute</summary>
[<RequireQualifiedAccess>]
module HxParams =
    
    /// <summary>Include all parameters</summary>
    let All = "*"
    
    /// <summary>Include no parameters</summary>
    let None = "none"
    
    /// <summary>Include the specified parameters</summary>
    /// <param name="fields">One or more fields to include in the request</param>
    /// <returns>The list of fields for the <tt>hx-params</tt> attribute value</returns>
    let With fields =
        match fields with [] -> "" | _ -> fields |> List.reduce (fun acc it -> $"{acc},{it}")
    
    /// <summary>Exclude the specified parameters</summary>
    /// <param name="fields">One or more fields to exclude from the request</param>
    /// <returns>The list of fields for the <tt>hx-params</tt> attribute value prefixed with "not"</returns>
    let Except fields =
        With fields |> sprintf "not %s"


/// <summary>Helpers to define <tt>hx-request</tt> attribute values</summary>
[<RequireQualifiedAccess>]
module HxRequest =
    
    open Giraffe.Htmx.Common
    
    /// <summary>Configure the request with various options</summary>
    /// <param name="opts">The options to configure</param>
    /// <returns>A string with the configured options</returns>
    let Configure (opts: string list) =
        opts
        |> String.concat ", "
        |> sprintf "{ %s }"
    
    /// <summary>Set a timeout (in milliseconds)</summary>
    /// <param name="ms">The milliseconds for the request timeout</param>
    /// <returns>A string with the configured request timeout</returns>
    let Timeout (ms: int) =
        $"\"timeout\": {ms}"
    
    /// <summary>Include or exclude credentials from the request</summary>
    /// <param name="send"><tt>true</tt> if credentials should be sent, <tt>false</tt> if not</param>
    /// <returns>A string with the configured credential options</returns>
    let Credentials send =
        (toLowerBool >> sprintf "\"credentials\": %s") send
    
    /// <summary>Exclude or include headers from the request</summary>
    /// <param name="exclude">
    /// <tt>true</tt> if no headers should be sent; <tt>false</tt> if headers should be sent
    /// </param>
    /// <returns>A string with the configured header options</returns>
    let NoHeaders exclude =
        (toLowerBool >> sprintf "\"noHeaders\": %s") exclude


/// Helpers for the `hx-trigger` attribute
[<RequireQualifiedAccess>]
module HxTrigger =
    
    /// Append a filter to a trigger
    let private appendFilter filter (trigger : string) =
        match trigger.Contains "[" with
        | true ->
            let parts = trigger.Split ('[', ']')
            $"{parts[0]}[{parts[1]}&&{filter}]"
        | false -> $"{trigger}[{filter}]"
    
    /// Trigger the event on a click
    let Click = "click"
    
    /// Trigger the event on page load
    let Load = "load"
    
    /// Trigger the event when the item is visible
    let Revealed = "revealed"
    
    /// Trigger this event every [timing declaration]
    let Every (duration : string) = $"every {duration}"
    
    /// Helpers for defining filters
    module Filter =

        /// Only trigger the event if the `ALT` key is pressed
        let Alt          = appendFilter "altKey"
        
        /// Only trigger the event if the `CTRL` key is pressed
        let Ctrl         = appendFilter "ctrlKey"
        
        /// Only trigger the event if the `SHIFT` key is pressed
        let Shift        = appendFilter "shiftKey"
        
        /// Only trigger the event if `CTRL+ALT` are pressed
        let CtrlAlt      = Ctrl    >> Alt
        
        /// Only trigger the event if `CTRL+SHIFT` are pressed
        let CtrlShift    = Ctrl    >> Shift
        
        /// Only trigger the event if `CTRL+ALT+SHIFT` are pressed
        let CtrlAltShift = CtrlAlt >> Shift
        
        /// Only trigger the event if `ALT+SHIFT` are pressed
        let AltShift     = Alt     >> Shift
    
    /// Append a modifier to the current trigger
    let private appendModifier modifier current =
        if current = "" then modifier else $"{current} {modifier}"
    
    /// Only trigger once
    let Once = appendModifier "once"
    
    /// Trigger when changed
    let Changed = appendModifier "changed"
    
    /// Delay execution; resets every time the event is seen
    let Delay = sprintf "delay:%s" >> appendModifier
    
    /// Throttle execution; ignore other events, fire when duration passes
    let Throttle = sprintf "throttle:%s" >> appendModifier
    
    /// Trigger this event from a CSS selector
    let From = sprintf "from:%s" >> appendModifier
    
    /// Trigger this event from the `document` object
    let FromDocument = From "document"
    
    /// Trigger this event from the `window` object
    let FromWindow = From "window"
    
    /// Trigger this event from the closest parent CSS selector
    let FromClosest = sprintf "closest %s" >> From
    
    /// Trigger this event from the closest child CSS selector
    let FromFind = sprintf "find %s" >> From
    
    /// Target the given CSS selector with the results of this event
    let Target = sprintf "target:%s" >> appendModifier
    
    /// Prevent any further events from occurring after this one fires
    let Consume = appendModifier "consume"
    
    /// Configure queueing when events fire when others are in flight; if unspecified, the default is "last"
    let Queue = sprintf "queue:%s" >> appendModifier
    
    /// Queue the first event, discard all others (i.e., a FIFO queue of 1)
    let QueueFirst = Queue "first"
    
    /// Queue the last event; discards current when another is received (i.e., a LIFO queue of 1)
    let QueueLast = Queue "last"
    
    /// Queue all events; discard none
    let QueueAll = Queue "all"
    
    /// Queue no events; discard all
    let QueueNone = Queue "none"


/// <summary>Helper to create the `hx-vals` attribute</summary>
[<RequireQualifiedAccess>]
module HxVals =
    
    /// <summary>Create values from a list of key/value pairs</summary>
    let From = Giraffe.Htmx.Common.toJson


/// <summary>Attributes and flags for htmx</summary>
[<AutoOpen>]
module HtmxAttrs =
    
    /// <summary>Progressively enhances anchors and forms to use AJAX requests</summary>
    /// <remarks>Use <tt>_hxNoBoost</tt> to set to false</remarks>
    let _hxBoost = attr "hx-boost" "true"
    
    /// <summary>Shows a <tt>confirm()</tt> dialog before issuing a request</summary>
    /// <param name="prompt">The prompt to present to the user when seeking their confirmation</param>
    /// <returns>A configured <tt>hx-confirm</tt> attribute</returns>
    let _hxConfirm prompt =
        attr "hx-confirm" prompt
    
    /// <summary>Issues a <tt>DELETE</tt> to the specified URL</summary>
    /// <param name="url">The URL to which the <tt>DELETE</tt> request should be sent</param>
    /// <returns>A configured <tt>hx-delete</tt> attribute</returns>
    let _hxDelete url =
        attr "hx-delete" url
    
    /// <summary>Disables htmx processing for the given node and any children nodes</summary>
    let _hxDisable = flag "hx-disable"
    
    /// <summary>Specifies elements that should be disabled when an htmx request is in flight</summary>
    /// <param name="elt">The element to disable when an htmx request is in flight</param>
    /// <returns>A configured <tt>hx-disabled-elt</tt> attribute</returns>
    let _hxDisabledElt elt =
        attr "hx-disabled-elt" elt

    /// <summary>Disinherit all ("*") or specific htmx attributes</summary>
    /// <param name="hxAttrs">The htmx attributes to disinherit (should start with "hx-")</param>
    /// <returns>A configured <tt>hx-disinherit</tt> attribute</returns>
    let _hxDisinherit hxAttrs =
        attr "hx-disinherit" hxAttrs
    
    /// <summary>Changes the request encoding type</summary>
    /// <param name="enc">The encoding type (use <tt>HxEncoding</tt> constants)</param>
    /// <returns>A configured <tt>hx-encoding</tt> attribute</returns>
    let _hxEncoding enc =
        attr "hx-encoding" enc
    
    /// <summary>Extensions to use for this element</summary>
    /// <param name="exts">A list of extensions to apply to this element</param>
    /// <returns>A configured <tt>hx-ext</tt> attribute</returns>
    let _hxExt exts =
        attr "hx-ext" exts
    
    /// <summary>Issues a <tt>GET</tt> to the specified URL</summary>
    /// <param name="url">The URL to which the <tt>GET</tt> request should be sent</param>
    /// <returns>A configured <tt>hx-get</tt> attribute</returns>
    let _hxGet url =
        attr "hx-get" url
    
    /// <summary>Adds to the headers that will be submitted with the request</summary>
    [<System.Obsolete "Convert this parameter">]
    let _hxHeaders    = attr "hx-headers"
    
    /// Set to "false" to prevent pages with sensitive information from being stored in the history cache
    let _hxHistory    = attr "hx-history"
    
    /// The element to snapshot and restore during history navigation
    let _hxHistoryElt = flag "hx-history-elt"
    
    /// Includes additional data in AJAX requests
    let _hxInclude    = attr "hx-include"
    
    /// The element to put the htmx-request class on during the AJAX request
    let _hxIndicator  = attr "hx-indicator"
    
    /// Overrides a previous `hx-boost`
    let _hxNoBoost    = attr "hx-boost" "false"
    
    /// Attach an event handler for DOM events
    let _hxOnEvent evtName =
        attr $"hx-on:%s{evtName}"

    /// Attach an event handler for htmx events
    let _hxOnHxEvent (hxEvent: HxEvent) =
        _hxOnEvent $":{hxEvent.ToHxOnString()}"

    /// Filters the parameters that will be submitted with a request
    let _hxParams     = attr "hx-params"
    
    /// Issues a PATCH to the specified URL
    let _hxPatch      = attr "hx-patch"
    
    /// Issues a POST to the specified URL
    let _hxPost       = attr "hx-post"
    
    /// Preserves an element between requests
    let _hxPreserve   = attr "hx-preserve" "true"
    
    /// Shows a prompt before submitting a request
    let _hxPrompt     = attr "hx-prompt"
    
    /// Pushes the URL into the location bar, creating a new history entry
    let _hxPushUrl    = attr "hx-push-url"
    
    /// Issues a PUT to the specified URL
    let _hxPut        = attr "hx-put"
    
    /// Replaces the current URL in the browser's history stack
    let _hxReplaceUrl = attr "hx-replace-url"
    
    /// Configures various aspects of the request
    let _hxRequest    = attr "hx-request"
    
    /// Selects a subset of the server response to process
    let _hxSelect     = attr "hx-select"
    
    /// Selects a subset of an out-of-band server response
    let _hxSelectOob  = attr "hx-select-oob"
    
    /// Controls how the response content is swapped into the DOM (e.g. 'outerHTML' or 'beforeEnd')
    let _hxSwap       = attr "hx-swap"
    
    /// Controls how the response content is swapped into the DOM (e.g. 'outerHTML' or 'beforeEnd'), enabling CSS
    /// transitions
    let _hxSwapWithTransition = sprintf "%s transition:true" >> _hxSwap
    
    /// Marks content in a response as being "Out of Band", i.e. swapped somewhere other than the target
    let _hxSwapOob    = attr "hx-swap-oob"
    
    /// Synchronize events based on another element
    let _hxSync       = attr "hx-sync"
    
    /// Specifies the target element to be swapped
    let _hxTarget     = attr "hx-target"
    
    /// Specifies the event that triggers the request
    let _hxTrigger    = attr "hx-trigger"
    
    /// Validate an input element (uses HTML5 validation API)
    let _hxValidate   = flag "hx-validate"
    
    /// Adds to the parameters that will be submitted with the request
    let _hxVals       = attr "hx-vals"
    
    /// The name of the message to swap into the DOM.
    let _sseSwap      = attr "sse-swap"

    /// The URL of the SSE server.
    let _sseConnect   = attr "sse-connect"


/// Script tags to pull htmx into a web page
module Script =
  
    /// Script tag to load the minified version from unpkg.com
    let minified =
        script [ _src         "https://unpkg.com/htmx.org@2.0.4"
                 _integrity   "sha384-HGfztofotfshcF7+8n44JQL2oJmowVChPTg48S+jvZoztPfvwD79OC/LTtG6dMp+"
                 _crossorigin "anonymous" ] []

    /// Script tag to load the unminified version from unpkg.com
    let unminified =
        script [ _src         "https://unpkg.com/htmx.org@2.0.4/dist/htmx.js"
                 _integrity   "sha384-oeUn82QNXPuVkGCkcrInrS1twIxKhkZiFfr2TdiuObZ3n3yIeMiqcRzkIcguaof1"
                 _crossorigin "anonymous" ] []


/// Functions to extract and render an HTML fragment from a document
[<RequireQualifiedAccess>]
module RenderFragment =
    
    /// Does this element have an ID matching the requested ID name?
    let private isIdElement nodeId (elt : XmlElement) =
        snd elt
        |> Array.exists (fun attr ->
            match attr with
            | KeyValue (name, value) -> name = "id" && value = nodeId
            | Boolean _ -> false)

    /// Generate a message if the requested ID node is not found
    let private nodeNotFound (nodeId : string) =
        $"<em>&ndash; ID {nodeId} not found &ndash;</em>"
    
    /// Find the node with the named ID
    let rec findIdNode nodeId (node : XmlNode) : XmlNode option =
        match node with
        | Text _ -> None
        | VoidElement elt -> if isIdElement nodeId elt then Some node else None
        | ParentNode (elt, children) ->
            if isIdElement nodeId elt then Some node else children |> List.tryPick (fun c -> findIdNode nodeId c)
    
    /// Functions to render a fragment as a string
    [<RequireQualifiedAccess>]
    module AsString =

        /// Render to HTML for the given ID
        let htmlFromNodes nodeId (nodes : XmlNode list) =
            match nodes |> List.tryPick(fun node -> findIdNode nodeId node) with
            | Some idNode -> RenderView.AsString.htmlNode idNode
            | None -> nodeNotFound nodeId

        /// Render to HTML for the given ID
        let htmlFromNode nodeId node =
            match findIdNode nodeId node with
            | Some idNode -> RenderView.AsString.htmlNode idNode
            | None -> nodeNotFound nodeId

    /// Functions to render a fragment as bytes
    [<RequireQualifiedAccess>]
    module AsBytes =

        let private utf8 = System.Text.Encoding.UTF8

        /// Render to HTML for the given ID
        let htmlFromNodes nodeId (nodes : XmlNode list) =
            match nodes |> List.tryPick(fun node -> findIdNode nodeId node) with
            | Some idNode -> RenderView.AsBytes.htmlNode idNode
            | None -> nodeNotFound nodeId |> utf8.GetBytes

        /// Render to HTML for the given ID
        let htmlFromNode nodeId node =
            match findIdNode nodeId node with
            | Some idNode -> RenderView.AsBytes.htmlNode idNode
            | None -> nodeNotFound nodeId |> utf8.GetBytes

    /// Functions to render a fragment into a StringBuilder
    [<RequireQualifiedAccess>]
    module IntoStringBuilder =

        /// Render to HTML for the given ID
        let htmlFromNodes sb nodeId (nodes : XmlNode list) =
            match nodes |> List.tryPick(fun node -> findIdNode nodeId node) with
            | Some idNode -> RenderView.IntoStringBuilder.htmlNode sb idNode
            | None -> nodeNotFound nodeId |> sb.Append |> ignore

        /// Render to HTML for the given ID
        let htmlFromNode sb nodeId node =
            match findIdNode nodeId node with
            | Some idNode -> RenderView.IntoStringBuilder.htmlNode sb idNode
            | None -> nodeNotFound nodeId |> sb.Append |> ignore
